#!/usr/bin/env python

import sys
import copy
import re
import numpy as np
import matplotlib.pyplot as plt

class Mesh():

    def __init__(self, directory=""):
        if len(directory) > 1:
            if directory[-1] != '/':
                directory += '/'
        try:
#            domain_x = np.loadtxt(directory+"domain_x.dat")
            domain_x = np.loadtxt("domain_x.dat")
        except IOError:
            print "IOError with domain_x.dat"
        try:
#            domain_y = np.loadtxt(directory+"domain_y.dat")[3:-3]
            domain_y = np.loadtxt("domain_y.dat")
        except IOError:
            print "IOError with domain_y.dat"
        self.xm = domain_x
        self.ym = domain_y

        self.xmed = 0.5*(domain_x[:-1] + domain_x[1:])
        self.ymed = 0.5*(domain_y[:-1] + domain_y[1:])

        T, R = np.meshgrid(self.xm, self.ym)
        R2  = R*R
        self.surf = 0.5*(T[:-1,1:]-T[:-1,:-1])*(R2[1:,:-1]-R2[:-1,:-1])

class Parameters():

    def __init__(self, directory=''):
        if len(directory) > 1:
            if directory[-1] != '/':
                directory += '/'
        paramfile = directory+"../in/in.par"
        try:
            params = open(paramfile,'r')
        except IOError:
            print  paramfile + " not found."
            return
        lines = params.readlines()
        params.close()
        par = {}
        for line in lines:
            if len(line.strip()) > 0 and line[0] != "#":
                l = line.split()
                name = l[0]
                value = l[1]
                try:
                    float(value)
                except ValueError:
                    try:
                        int(value)
                    except ValueError:
                        value = '"' + value + '"'
                par[name] = value
        self._params = par
        for name in par:
            exec("self." + name.lower() + "=" + par[name])

class Field(Mesh, Parameters):

    def __init__(self, field, staggered='c', directory='', dtype='float64'):
        if len(directory) > 1:
            if directory[-1] != '/':
                directory += '/'
        Mesh.__init__(self, directory)
        Parameters.__init__(self, directory+'../in/')

        self.nx = self.nsec
        self.ny = self.nrad
        self.xmin = 0.0
        self.xmax = 2.0*np.pi
        self.ymin = self.rmin
        self.ymax = self.rmax

        if staggered.count('x')>0:
            self.x = self.xm[:-1]
        else:
            self.x = self.xmed
        if staggered.count('y')>0:
            self.y = self.ym[:-1]
        else:
            self.y = self.ymed

        self.data = self.__open_field(directory+field, dtype)

    def __open_field(self, f, dtype):
        print "file = ", f
        field = np.fromfile(f, dtype=dtype)
        return field.reshape(self.ny, self.nx)

    def plot(self, log=False, cartesian=False, cmap='magma', usemean=False, **karg):
        ax = plt.gca()
        if log:
            data = np.log(self.data)
        else:
            data = self.data

        if usemean:
            mean1 = np.average(data, axis=1)
            mean = np.zeros((self.ny-1,self.nx-1))
            for j in xrange(0,self.nx-1):
                mean[:,j] = mean1
            data = (data-mean)/mean

        if cartesian:
            T, R = np.meshgrid(self.x, self.y)
            X = R*np.cos(T)
            Y = R*np.sin(T)
            ap = plt.pcolormesh(X, Y, data, cmap=cmap, **karg)
            ax.set_aspect('equal')
        else:
#            ap = ax.imshow(data, cmap = cmap, origin='lower',aspect='auto', extent=[self.x[0],self.x[-1],self.y[0],self.y[-1]], **karg)
            dx = self.x[1]-self.x[0]
            dy = self.y[1]-self.y[0]
            ap = ax.imshow(data, cmap = cmap, origin='lower',aspect='auto', extent=[self.x[0]-0.5*dx,self.x[-1]+0.5*dx,self.y[0]-0.5*dy,self.y[-1]+0.5*dy], **karg)
        return ap

    def contour(self, log=False, cartesian=False, **karg):
        if log:
            data = np.log(self.data)
        else:
            data = self.data
        ax = plt.gca()
        T, R = np.meshgrid(self.x, self.y)
        if cartesian:
            X = R*np.cos(T)
            Y = R*np.sin(T)
            plt.contour(X, Y, data, **karg)
        else:
            plt.contour(T, R, data, **karg)

def shift_field(Field, direction):
    F = copy.deepcopy(Field)
    if direction.count('x')>0:
        F.data = 0.5*(Field.data[:,1:]+Field.data[:,:-1])
        F.x = 0.5*(Field.x[1:]+Field.x[:-1])
    if direction.count('y')>0:
        F.data = 0.5*(F.data[1:,:]+F.data[:-1,:])
        F.y = 0.5*(F.y[1:]+F.y[:-1])

    F.nx = len(F.x)
    F.ny = len(F.y)

    return F

def cut_field(Field, direction, side):
    _cut_field = copy.deepcopy(Field)
    ny,nx = Field.ny, Field.nx
    mx = my = px = py = 0

    if direction.count('x')>0:
        if side.count('m')>0:
            mx = 1
        if side.count('p')>0:
            px = 1
    if direction.count('y')>0:
        if side.count('m')>0:
            my = 1
        if side.count('p')>0:
            py = 1

    _cut_field.data = Field.data[my:ny-py,mx:nx-px]
    _cut_field.x = _cut_field.x[mx:nx-px]
    _cut_field.y = _cut_field.y[my:ny-py]

    return _cut_field

def vector_field(vx, vy, scale=0.2, **karg):
    nsx = nsy = 3
    T, R = np.meshgrid(vx.x[::nsx], vx.y[::nsy])
    X = R*np.cos(T)
    Y = R*np.sin(T)
    vx = vx.data[::nsy,::nsx]
    vy = vy.data[::nsy,::nsx]
    U = vy*np.cos(T) - vx*np.sin(T)
    V = vy*np.sin(T) + vx*np.cos(T)
    ax = plt.gca()
    ax.quiver(X, Y, U, V, scale=scale, pivot='middle', **karg)

def euler(vx, vy, x, y, reverse):
    sign = 1.0
    if reverse:
        sign = -1
    vphi = get_v(vx,x,y)
    vrad = get_v(vy,x,y)
    if vphi == None or vrad == None:
        return None, None
    l = np.min((((vx.xmax-vx.xmin)/vx.nx),((vx.ymax-vx.ymin)/vx.ny)))
    h = 0.5*l/np.sqrt((vphi**2+vrad**2))

    return sign*h*np.array([vphi/y,vrad])

def bilinear(x, y, f, p):
    xp  = p[0]; yp  = p[1]
    x1  = x[0]; x2  = x[1]
    y1  = y[0]; y2  = y[1]
    f11 = f[0]; f12 = f[1]; f21 = f[2]; f22 = f[3]
    t = (xp-x1)/(x2-x1)
    u = (yp-y1)/(y2-y1)

    return (1.0-t)*(1.0-u)*f11 + t*(1.0-u)*f12 + t*u*f22 + u*(1-t)*f21

def get_v(v, x, y):
    i = int((x-v.xmin)/(v.xmax-v.xmin)*v.nx)
    j = int((y-v.ymin)/(v.ymax-v.ymin)*v.ny)

    if i<0 or j<0 or i>v.data.shape[1]-2 or j>v.data.shape[0]-2:
        return None

    f11 = v.data[j,i]
    f12 = v.data[j,i+1]
    f21 = v.data[j+1,i]
    f22 = v.data[j+1,i+1]
    try:
        x1  = v.x[i]
        x2  = v.x[i+1]
        y1  = v.y[j]
        y2  = v.y[j+1]
        return bilinear((x1,x2),(y1,y2),(f11,f12,f21,f22),(x,y))
    except IndexError:
        return None

def get_stream(vx, vy, x0, y0, nmax=1000000, maxlength=4*np.pi, bidirectional=True, reverse=False):
    if bidirectional:
        s0 = get_stream(vx, vy, x0, y0, reverse=False, bidirectional=False, nmax=nmax,maxlength=maxlength)
        s1 = get_stream(vx, vy, x0, y0, reverse=True,  bidirectional=False, nmax=nmax,maxlength=maxlength)
        return (s0,s1)

    l = 0
    x = [x0]
    y = [y0]

    for i in xrange(nmax):
        ds = euler(vx, vy, x0, y0, reverse=reverse)
        if ds[0] == None:
            if len(x)==1:
                print "There was an error getting the stream, ds is NULL (see get_stream)."
                print "x0 = ", x0
                print "y0 = ", y0
                print "ds = ", ds
            break
        l += np.sqrt(ds[0]**2+ds[1]**2)
        dx = ds[0]
        dy = ds[1]
        if(np.sqrt(dx**2+dy**2)<1e-13):
            print "Warning (get_stream): ds is very small, check if you're in a stagnation point."
            print "Try selecting another initial point."
            break
        if l > maxlength:
            print "maxlength reached: ", l
            break
        x0 += dx
        y0 += dy
        x.append(x0)
        y.append(y0)

    return np.array([x,y])

def get_random_streams(vx, vy, xmin=None, xmax=None, ymin=None, ymax=None, n=30, nmax=100000):
    if xmin == None:
        xmin = vx.xmin
    if ymin == None:
        ymin = vx.ymin
    if xmax == None:
        xmax = vx.xmax
    if ymax == None:
        ymax = vx.ymax
    X = xmin + np.random.rand(n)*(xmax-xmin)
    Y = ymin + np.random.rand(n)*(ymax-ymin)
    streams = []
    counter = 0
    for x,y in zip(X,Y):
        stream = get_stream(vx, vy, x, y, nmax=nmax, bidirectional=True)
        streams.append(stream)
        counter += 1
    return streams

def plot_streams(streams, cartesian=False, **kargs):
    ax = plt.gca()
    for stream in streams:
        for sub_stream in stream:
            if cartesian:
                ax.plot(sub_stream[1]*np.cos(sub_stream[0]), sub_stream[1]*np.sin(sub_stream[0]), **kargs)
            else:
                ax.plot(sub_stream[0], sub_stream[1], **kargs)

def streamline(filename, directory='', out=None, interactive=False, logscale=False, c1=None, c2=None,
    xp=None, yp=None, zp=None, vxp=None, vyp=None, vzp=None, Rp=None, RH=None, ip=None, center=None):

    param = Parameters(directory)
    print "param.nrad = ", param.nrad
    print "param.nsec = ", param.nsec

    x = np.linspace(0, 2.0*np.pi, param.nsec+1)
    np.savetxt("domain_x.dat", x)

    y = np.linspace(param.rmin, param.rmax, param.nrad+1)
    np.savetxt("domain_y.dat", y)

    rho = Field(filename, directory=directory)
    vx = Field(filename.replace("gasdens","gasvtheta"), directory=directory, staggered='x')
    vy = Field(filename.replace("gasdens","gasvrad"), directory=directory, staggered='y')

    vxc = shift_field(vx,'x')
    vyc = shift_field(vy,'y')

    vxcc = cut_field(vxc, direction='y', side='p')
    vycc = cut_field(vyc, direction='x', side='p')
    rhoc = cut_field(rho, direction='xy', side='p')

    if xp != None:
        xc = xp[center]
        yc = yp[center]
        zc = zp[center]
        rc = np.sqrt(xc**2+yc**2+zc**2)
    else:
        rc = 0.723183950123
    if vxp != None:
        vxc = vxp[center]
        vyc = vyp[center]
        vzc = vzp[center]
        vc = np.sqrt(vxc**2+vyc**2+vzc**2)
    else:
        vc = rc**(-0.5)

    for j in xrange(0,vxcc.ny-1):
        vxcc.data[j,:] -= vc

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_xlabel("$\phi$ [rad]")
    ax.set_ylabel("$r$ [au]")
    ax.set_xlim(0.0,2.0*np.pi)
    tmp = 0.15
    ax.set_ylim(rc-tmp,rc+tmp)

    ap = rhoc.plot(usemean=True)
    cb = plt.colorbar(ap)
    cb.set_label("$(\Sigma-\Sigma(r))/\Sigma(r)$")
    if c1 != None:
        ap.set_clim(c1,c2)

#    vector_field(vxcc, vycc, color='r')

    if xp != None:
        for i in xrange(0, len(xp)):
            r = np.sqrt(xp[i]*xp[i]+yp[i]*yp[i])
            phi = np.arctan2(yp[i],xp[i])
            if phi < 0.0:
                phi += 2.0*np.pi
            if xc != None and (r > rc-tmp and r < rc+tmp):
                plt.plot(phi, r, 'w+', markersize=3, markerfacecolor='none', markeredgecolor='white', markeredgewidth=0.2, alpha=0.5)

    tmp = 0.02
    np.random.seed(1)
    streams = get_random_streams(vxcc, vycc, n=50, xmin=0.0, xmax=2.0*np.pi, ymin=rc-tmp, ymax=rc+tmp)
    plot_streams(streams, cartesian=False, color='w', linewidth=0.2, alpha=0.5)

    if out != None:
        plt.savefig(out, dpi=300, bbox_inches='tight', pad_inches=0)
    if interactive:
        plt.show()

def main():
    streamline("gasdens4.dat", out="streamline.png")

if __name__ == "__main__":
    main()


